package java.util;

import java.io.IOException;
import java.io.FilterOutputStream;
import java.io.InputStream;
import java.nio.charset.StandardCharsets;
import java.io.OutputStream;
import java.nio.ByteBuffer;

public class Base64
{
    private Base64() {
    }
    
    public static Base64.Encoder getEncoder() {
        return Base64.Encoder.RFC4648;
    }
    
    public static Base64.Encoder getUrlEncoder() {
        return Base64.Encoder.RFC4648_URLSAFE;
    }
    
    public static Base64.Encoder getMimeEncoder() {
        return Base64.Encoder.RFC2045;
    }
    
    public static Base64.Encoder getMimeEncoder(int n, byte[] array) {
        Objects.requireNonNull(array);
        int[] access$000 = Base64.Decoder.fromBase64;
        int length = array.length;
        int i = 0;
        while (i < length) {
            int j = array[i];
            if (access$000[j & 0xFF] != -1) {
                throw new IllegalArgumentException("Illegal base64 line separator character 0x" + Integer.toString(j, 16));
            }
            ++i;
        }
        if (n <= 0) {
            return Base64.Encoder.RFC4648;
        }
        return new Base64.Encoder(false, array, n >> 2 << 2, true);
    }
    
    public static Base64.Decoder getDecoder() {
        return Base64.Decoder.RFC4648;
    }
    
    public static Base64.Decoder getUrlDecoder() {
        return Base64.Decoder.RFC4648_URLSAFE;
    }
    
    public static Base64.Decoder getMimeDecoder() {
        return Base64.Decoder.RFC2045;
    }
    
    public static class Encoder
    {
        private final byte[] newline;
        private final int linemax;
        private final boolean isURL;
        private final boolean doPadding;
        private static final char[] toBase64;
        private static final char[] toBase64URL;
        private static final int MIMELINEMAX = 76;
        private static final byte[] CRLF;
        static final Base64.Encoder RFC4648;
        static final Base64.Encoder RFC4648_URLSAFE;
        static final Base64.Encoder RFC2045;
        
        private Encoder(boolean isURL, byte[] newline, int linemax, boolean doPadding) {
            this.isURL = isURL;
            this.newline = newline;
            this.linemax = linemax;
            this.doPadding = doPadding;
        }
        
        private final int outLength(int n) {
            int n3;
            if (this.doPadding) {
                n3 = 4 * ((n + 2) / 3);
            }
            else {
                int n2 = n % 3;
                n3 = 4 * (n / 3) + ((n2 == 0) ? 0 : (n2 + 1));
            }
            if (this.linemax > 0) {
                n3 += (n3 - 1) / this.linemax * this.newline.length;
            }
            return n3;
        }
        
        public byte[] encode(byte[] array) {
            byte[] array2 = new byte[this.outLength(array.length)];
            int encode0 = this.encode0(array, 0, array.length, array2);
            if (encode0 != array2.length) {
                return Arrays.copyOf(array2, encode0);
            }
            return array2;
        }
        
        public int encode(byte[] array, byte[] array2) {
            if (array2.length < this.outLength(array.length)) {
                throw new IllegalArgumentException("Output byte array is too small for encoding all input bytes");
            }
            return this.encode0(array, 0, array.length, array2);
        }
        
        public String encodeToString(byte[] array) {
            byte[] encode = this.encode(array);
            return new String(encode, 0, 0, encode.length);
        }
        
        public ByteBuffer encode(ByteBuffer byteBuffer) {
            byte[] copy = new byte[this.outLength(byteBuffer.remaining())];
            int n;
            if (byteBuffer.hasArray()) {
                n = this.encode0(byteBuffer.array(), byteBuffer.arrayOffset() + byteBuffer.position(), byteBuffer.arrayOffset() + byteBuffer.limit(), copy);
                byteBuffer.position(byteBuffer.limit());
            }
            else {
                byte[] dst = new byte[byteBuffer.remaining()];
                byteBuffer.get(dst);
                n = this.encode0(dst, 0, dst.length, copy);
            }
            if (n != copy.length) {
                copy = Arrays.copyOf(copy, n);
            }
            return ByteBuffer.wrap(copy);
        }
        
        public OutputStream wrap(OutputStream outputStream) {
            Objects.requireNonNull(outputStream);
            return new Base64.EncOutputStream(outputStream, this.isURL ? Base64.Encoder.toBase64URL : Base64.Encoder.toBase64, this.newline, this.linemax, this.doPadding);
        }
        
        public Base64.Encoder withoutPadding() {
            if (!this.doPadding) {
                return this;
            }
            return new Base64.Encoder(this.isURL, this.newline, this.linemax, false);
        }
        
        private int encode0(byte[] array, int i, int n, byte[] array2) {
            char[] array3 = this.isURL ? Base64.Encoder.toBase64URL : Base64.Encoder.toBase64;
            int n2 = (n - i) / 3 * 3;
            int b = i + n2;
            if (this.linemax > 0 && n2 > this.linemax / 4 * 3) {
                n2 = this.linemax / 4 * 3;
            }
            int n3 = 0;
            while (i < b) {
                int min;
                int n4;
                for (min = Math.min(i + n2, b), , ; i < min; n4 = ((array[i++] & 0xFF) << 16 | (array[i++] & 0xFF) << 8 | (array[i++] & 0xFF)), array2[n3++] = (byte)array3[n4 >>> 18 & 0x3F], array2[n3++] = (byte)array3[n4 >>> 12 & 0x3F], array2[n3++] = (byte)array3[n4 >>> 6 & 0x3F], array2[n3++] = (byte)array3[n4 & 0x3F]) {}
                int n5 = (min - i) / 3 * 4;
                n3 += n5;
                i = min;
                if (n5 == this.linemax && i < n) {
                    byte[] newline = this.newline;
                    for (int length = newline.length, j = 0; j < length; ++j) {
                        array2[n3++] = newline[j];
                    }
                }
                else {
                    continue;
                }
            }
            if (i < n) {
                int n6 = array[i++] & 0xFF;
                array2[n3++] = (byte)array3[n6 >> 2];
                if (i == n) {
                    array2[n3++] = (byte)array3[n6 << 4 & 0x3F];
                    if (this.doPadding) {
                        array2[n3++] = 61;
                        array2[n3++] = 61;
                    }
                }
                else {
                    int n7 = array[i++] & 0xFF;
                    array2[n3++] = (byte)array3[(n6 << 4 & 0x3F) | n7 >> 4];
                    array2[n3++] = (byte)array3[n7 << 2 & 0x3F];
                    if (this.doPadding) {
                        array2[n3++] = 61;
                    }
                }
            }
            return n3;
        }
        
        static {
            toBase64 = new char[] { 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z', 'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z', '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '+', '/' };
            toBase64URL = new char[] { 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z', 'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z', '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '-', '_' };
            CRLF = new byte[] { 13, 10 };
            RFC4648 = new Base64.Encoder(false, null, -1, true);
            RFC4648_URLSAFE = new Base64.Encoder(true, null, -1, true);
            RFC2045 = new Base64.Encoder(false, Base64.Encoder.CRLF, 76, true);
        }
    }
    
    public static class Decoder
    {
        private final boolean isURL;
        private final boolean isMIME;
        private static final int[] fromBase64;
        private static final int[] fromBase64URL;
        static final Base64.Decoder RFC4648;
        static final Base64.Decoder RFC4648_URLSAFE;
        static final Base64.Decoder RFC2045;
        
        private Decoder(boolean isURL, boolean isMIME) {
            this.isURL = isURL;
            this.isMIME = isMIME;
        }
        
        public byte[] decode(byte[] array) {
            byte[] copy = new byte[this.outLength(array, 0, array.length)];
            int decode0 = this.decode0(array, 0, array.length, copy);
            if (decode0 != copy.length) {
                copy = Arrays.copyOf(copy, decode0);
                return copy;
            }
            return copy;
        }
        
        public byte[] decode(String s) {
            return this.decode(s.getBytes(StandardCharsets.ISO_8859_1));
        }
        
        public int decode(byte[] array, byte[] array2) {
            if (array2.length < this.outLength(array, 0, array.length)) {
                throw new IllegalArgumentException("Output byte array is too small for decoding all input bytes");
            }
            return this.decode0(array, 0, array.length, array2);
        }
        
        public ByteBuffer decode(ByteBuffer byteBuffer) {
            int position = byteBuffer.position();
            try {
                byte[] array;
                int n;
                int length;
                if (byteBuffer.hasArray()) {
                    array = byteBuffer.array();
                    n = byteBuffer.arrayOffset() + byteBuffer.position();
                    length = byteBuffer.arrayOffset() + byteBuffer.limit();
                    byteBuffer.position(byteBuffer.limit());
                }
                else {
                    array = new byte[byteBuffer.remaining()];
                    byteBuffer.get(array);
                    n = 0;
                    length = array.length;
                }
                byte[] array2 = new byte[this.outLength(array, n, length)];
                return ByteBuffer.wrap(array2, 0, this.decode0(array, n, length, array2));
            }
            catch (IllegalArgumentException ex) {
                byteBuffer.position(position);
                throw ex;
            }
        }
        
        public InputStream wrap(InputStream inputStream) {
            Objects.requireNonNull(inputStream);
            return new Base64.DecInputStream(inputStream, this.isURL ? Base64.Decoder.fromBase64URL : Base64.Decoder.fromBase64, this.isMIME);
        }
        
        private int outLength(byte[] array, int i, int n) {
            int[] array2 = this.isURL ? Base64.Decoder.fromBase64URL : Base64.Decoder.fromBase64;
            int n2 = 0;
            int n3 = n - i;
            if (n3 == 0) {
                return 0;
            }
            if (n3 < 2) {
                if (this.isMIME && array2[0] == -1) {
                    return 0;
                }
                throw new IllegalArgumentException("Input byte[] should at least have 2 bytes for base64 bytes");
            }
            else {
                if (this.isMIME) {
                    int n4 = 0;
                    while (i < n) {
                        int n5 = array[i++] & 0xFF;
                        if (n5 == 61) {
                            n3 -= n - i + 1;
                            break;
                        }
                        if (array2[n5] == -1) {
                            ++n4;
                        }
                        else {
                            continue;
                        }
                    }
                    n3 -= n4;
                }
                else if (array[n - 1] == 61) {
                    ++n2;
                    if (array[n - 2] == 61) {
                        ++n2;
                    }
                }
                if (n2 == 0 && (n3 & 0x3) != 0x0) {
                    n2 = 4 - (n3 & 0x3);
                }
                return 3 * ((n3 + 3) / 4) - n2;
            }
        }
        
        private int decode0(byte[] array, int i, int n, byte[] array2) {
            int[] array3 = this.isURL ? Base64.Decoder.fromBase64URL : Base64.Decoder.fromBase64;
            int n2 = 0;
            int n3 = 0;
            int n4 = 18;
            while (i < n) {
                int n5;
                if ((n5 = array3[array[i++] & 0xFF]) < 0) {
                    if (n5 == -2) {
                        if ((n4 == 6 && (i == n || array[i++] != 61)) || n4 == 18) {
                            throw new IllegalArgumentException("Input byte array has wrong 4-byte ending unit");
                        }
                        break;
                    }
                    else {
                        if (this.isMIME) {
                            continue;
                        }
                        throw new IllegalArgumentException("Illegal base64 character " + Integer.toString((int)array[i - 1], 16));
                    }
                }
                else {
                    n3 |= n5 << n4;
                    n4 -= 6;
                    if (n4 < 0) {
                        array2[n2++] = (byte)(n3 >> 16);
                        array2[n2++] = (byte)(n3 >> 8);
                        array2[n2++] = (byte)n3;
                        n4 = 18;
                        n3 = 0;
                    }
                    else {
                        continue;
                    }
                }
            }
            if (n4 == 6) {
                array2[n2++] = (byte)(n3 >> 16);
            }
            else if (n4 == 0) {
                array2[n2++] = (byte)(n3 >> 16);
                array2[n2++] = (byte)(n3 >> 8);
            }
            else if (n4 == 12) {
                throw new IllegalArgumentException("Last unit does not have enough valid bits");
            }
            while (i < n) {
                if (this.isMIME && array3[array[i++]] < 0) {
                    continue;
                }
                throw new IllegalArgumentException("Input byte array has incorrect ending byte at " + i);
            }
            return n2;
        }
        
        static {
            Arrays.fill(fromBase64 = new int[256], -1);
            for (int i = 0; i < Base64.Encoder.toBase64.length; ++i) {
                Base64.Decoder.fromBase64[Base64.Encoder.toBase64[i]] = i;
            }
            Base64.Decoder.fromBase64[61] = -2;
            Arrays.fill(fromBase64URL = new int[256], -1);
            for (int j = 0; j < Base64.Encoder.toBase64URL.length; ++j) {
                Base64.Decoder.fromBase64URL[Base64.Encoder.toBase64URL[j]] = j;
            }
            Base64.Decoder.fromBase64URL[61] = -2;
            RFC4648 = new Base64.Decoder(false, false);
            RFC4648_URLSAFE = new Base64.Decoder(true, false);
            RFC2045 = new Base64.Decoder(false, true);
        }
    }
    
    private static class EncOutputStream extends FilterOutputStream
    {
        private int leftover;
        private int b0;
        private int b1;
        private int b2;
        private boolean closed;
        private final char[] base64;
        private final byte[] newline;
        private final int linemax;
        private final boolean doPadding;
        private int linepos;
        
        EncOutputStream(OutputStream out, char[] base64, byte[] newline, int linemax, boolean doPadding) {
            super(out);
            this.leftover = 0;
            this.closed = false;
            this.linepos = 0;
            this.base64 = base64;
            this.newline = newline;
            this.linemax = linemax;
            this.doPadding = doPadding;
        }
        
        @Override
        public void write(int n) throws IOException {
            this.write(new byte[] { (byte)(n & 0xFF) }, 0, 1);
        }
        
        private void checkNewline() throws IOException {
            if (this.linepos == this.linemax) {
                this.out.write(this.newline);
                this.linepos = 0;
            }
        }
        
        @Override
        public void write(byte[] array, int n, int n2) throws IOException {
            if (this.closed) {
                throw new IOException("Stream is closed");
            }
            if (n < 0 || n2 < 0 || n2 > array.length - n) {
                throw new ArrayIndexOutOfBoundsException();
            }
            if (n2 == 0) {
                return;
            }
            if (this.leftover != 0) {
                if (this.leftover == 1) {
                    this.b1 = (array[n++] & 0xFF);
                    if (--n2 == 0) {
                        ++this.leftover;
                        return;
                    }
                }
                this.b2 = (array[n++] & 0xFF);
                --n2;
                this.checkNewline();
                this.out.write((int)this.base64[this.b0 >> 2]);
                this.out.write((int)this.base64[(this.b0 << 4 & 0x3F) | this.b1 >> 4]);
                this.out.write((int)this.base64[(this.b1 << 2 & 0x3F) | this.b2 >> 6]);
                this.out.write((int)this.base64[this.b2 & 0x3F]);
                this.linepos += 4;
            }
            int n3 = n2 / 3;
            this.leftover = n2 - n3 * 3;
            while (n3-- > 0) {
                this.checkNewline();
                int n4 = (array[n++] & 0xFF) << 16 | (array[n++] & 0xFF) << 8 | (array[n++] & 0xFF);
                this.out.write((int)this.base64[n4 >>> 18 & 0x3F]);
                this.out.write((int)this.base64[n4 >>> 12 & 0x3F]);
                this.out.write((int)this.base64[n4 >>> 6 & 0x3F]);
                this.out.write((int)this.base64[n4 & 0x3F]);
                this.linepos += 4;
            }
            if (this.leftover == 1) {
                this.b0 = (array[n++] & 0xFF);
            }
            else if (this.leftover == 2) {
                this.b0 = (array[n++] & 0xFF);
                this.b1 = (array[n++] & 0xFF);
            }
            return;
        }
        
        @Override
        public void close() throws IOException {
            if (!this.closed) {
                this.closed = true;
                if (this.leftover == 1) {
                    this.checkNewline();
                    this.out.write((int)this.base64[this.b0 >> 2]);
                    this.out.write((int)this.base64[this.b0 << 4 & 0x3F]);
                    if (this.doPadding) {
                        this.out.write(61);
                        this.out.write(61);
                    }
                }
                else if (this.leftover == 2) {
                    this.checkNewline();
                    this.out.write((int)this.base64[this.b0 >> 2]);
                    this.out.write((int)this.base64[(this.b0 << 4 & 0x3F) | this.b1 >> 4]);
                    this.out.write((int)this.base64[this.b1 << 2 & 0x3F]);
                    if (this.doPadding) {
                        this.out.write(61);
                    }
                }
                this.leftover = 0;
                this.out.close();
            }
        }
    }
    
    private static class DecInputStream extends InputStream
    {
        private final InputStream is;
        private final boolean isMIME;
        private final int[] base64;
        private int bits;
        private int nextin;
        private int nextout;
        private boolean eof;
        private boolean closed;
        private byte[] sbBuf;
        
        DecInputStream(InputStream is, int[] base64, boolean isMIME) {
            this.bits = 0;
            this.nextin = 18;
            this.nextout = -8;
            this.eof = false;
            this.closed = false;
            this.sbBuf = new byte[1];
            this.is = is;
            this.base64 = base64;
            this.isMIME = isMIME;
        }
        
        @Override
        public int read() throws IOException {
            return (this.read(this.sbBuf, 0, 1) == -1) ? -1 : (this.sbBuf[0] & 0xFF);
        }
        
        @Override
        public int read(byte[] array, int n, int i) throws IOException {
            if (this.closed) {
                throw new IOException("Stream is closed");
            }
            if (this.eof && this.nextout < 0) {
                return -1;
            }
            if (n < 0 || i < 0 || i > array.length - n) {
                throw new IndexOutOfBoundsException();
            }
            Label_0118: {
                if (this.nextout >= 0) {
                    while (i != 0) {
                        array[n++] = (byte)(this.bits >> this.nextout);
                        --i;
                        this.nextout -= 8;
                        if (this.nextout >= 0) {
                            continue;
                        }
                        this.bits = 0;
                        break Label_0118;
                    }
                    return n - n;
                }
            }
            while (i > 0) {
                int read = this.is.read();
                if (read == -1) {
                    this.eof = true;
                    if (this.nextin != 18) {
                        if (this.nextin == 12) {
                            throw new IOException("Base64 stream has one un-decoded dangling byte.");
                        }
                        array[n++] = (byte)(this.bits >> 16);
                        --i;
                        if (this.nextin == 0) {
                            if (i == 0) {
                                this.bits >>= 8;
                                this.nextout = 0;
                            }
                            else {
                                array[n++] = (byte)(this.bits >> 8);
                            }
                        }
                    }
                    if (n == n) {
                        return -1;
                    }
                    return n - n;
                }
                else if (read == 61) {
                    if (this.nextin == 18 || this.nextin == 12 || (this.nextin == 6 && this.is.read() != 61)) {
                        throw new IOException("Illegal base64 ending sequence:" + this.nextin);
                    }
                    array[n++] = (byte)(this.bits >> 16);
                    --i;
                    if (this.nextin == 0) {
                        if (i == 0) {
                            this.bits >>= 8;
                            this.nextout = 0;
                        }
                        else {
                            array[n++] = (byte)(this.bits >> 8);
                        }
                    }
                    this.eof = true;
                    break;
                }
                else {
                    int j;
                    if ((j = this.base64[read]) == -1) {
                        if (this.isMIME) {
                            continue;
                        }
                        throw new IOException("Illegal base64 character " + Integer.toString(j, 16));
                    }
                    else {
                        this.bits |= j << this.nextin;
                        if (this.nextin == 0) {
                            this.nextin = 18;
                            this.nextout = 16;
                            while (this.nextout >= 0) {
                                array[n++] = (byte)(this.bits >> this.nextout);
                                --i;
                                this.nextout -= 8;
                                if (i != 0 || this.nextout < 0) {
                                    continue;
                                }
                                return n - n;
                            }
                            this.bits = 0;
                        }
                        else {
                            this.nextin -= 6;
                        }
                    }
                }
            }
            return n - n;
        }
        
        @Override
        public int available() throws IOException {
            if (this.closed) {
                throw new IOException("Stream is closed");
            }
            return this.is.available();
        }
        
        @Override
        public void close() throws IOException {
            if (!this.closed) {
                this.closed = true;
                this.is.close();
            }
        }
    }
}
